import type {
  UserListResponseModel,
  ListViewConfig,
  AddViewConfig,
  UserDetailResponseModel,
  LabeledValue,
  FormValues,
  ToggleActiveAction,
  ToggleWifiActiveAction,
} from '@/helpers/models';
import {fetchAuthenticated} from '@/helpers/retrieveData';
import {useErrorStore} from '@/stores/error';
import i18next from 'i18next';
import {
  addViewConfig,
  listViewConfig,
  fetchMockData,
  csvResponseModel,
  getLabeledGroups,
  setPasswordResponseModel,
  userDetailResponseModel,
  getUserListResponseModels,
  setWifiPasswordResponseModel,
} from '@/helpers/mocks';

const needMock = import.meta.env.VITE_USE_REAL_BACKEND !== 'true';

const getError = async(response: Response): Promise<{detail: unknown} | null> => {
  if (!response.ok) {
    const text = await response.text() || response.statusText;
    let detail = {
      detail: text,
    };
    try {
      const json = JSON.parse(text);
      if (json && typeof json === 'object' && 'detail' in json) {
        detail = json;
      }
    } catch (err) {
      // Either not a JSON response, or unparsable. We'll just ignore this.
    }
    return detail;
  }
  return null;
};

export const fetchListViewConfig = async(): Promise<ListViewConfig> => {
  if (needMock) {
    return fetchMockData(listViewConfig, 'fetchListViewConfig');
  }

  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/pageconf/listView');
  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchListViewConfig.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
      unRecoverable: true,
    });
    return Promise.reject(answer);
  }
  return await answer.json();
};

export const fetchAddViewConfig = async(): Promise<AddViewConfig> => {
  if (needMock) {
    return fetchMockData(addViewConfig, 'fetchAddViewConfig');
  }
  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/pageconf/addView');
  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchListViewConfig.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }

  return await answer.json();
};

export const fetchUsers = async(
  school: string,
  group: string,
  role: string,
  sourceUID: string,
  state: string,
  quickSearch: string,
  networkAccess: string,
  limit?: number,
): Promise<UserListResponseModel[] | null> => {
  if (needMock) {
    return fetchMockData(getUserListResponseModels(), 'fetchUsers');
  }
  const urlSearchParams = new URLSearchParams({school});
  if (group !== '') {
    urlSearchParams.append('group', group);
  }
  if (role !== '') {
    urlSearchParams.append('role', role);
  }
  if (sourceUID !== '') {
    urlSearchParams.append('sourceUID', sourceUID);
  }
  if (quickSearch !== '') {
    urlSearchParams.append('quickSearch', quickSearch);
  }
  if (state !== '') {
    urlSearchParams.append('disabled', (state === 'disabled').toString());
  }
  if (networkAccess) {
    urlSearchParams.append('networkAccess', (networkAccess === 'enabled').toString());
  }
  if (limit) {
    urlSearchParams.append('limit', `${limit}`);
  }

  const answer = await fetchAuthenticated(`/ucsschool/bff-users/v1/users?${urlSearchParams.toString()}`);
  const error = await getError(answer);
  if (error) {
    if (typeof error.detail === 'object' && error.detail !== null && 'type' in error.detail) {
      const detail = error.detail as {type: string};
      if (detail.type === 'SearchLimitReached') {
        return null;
      }
    }
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchUsers.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }
  return await answer.json();
};

export const fetchGroups = async(school: string, groupKind: string, addGroupKindSuffixToLabel = false): Promise<LabeledValue<string>[]> => {
  if (needMock) {
    return fetchMockData(getLabeledGroups(), 'fetchGroups');
  }

  const urlSearchParams = new URLSearchParams({school});
  const answer = await fetchAuthenticated(`/ucsschool/bff-users/v1/${groupKind}?${urlSearchParams.toString()}`);
  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchGroups.error.title', {
      context: groupKind,
    });
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }
  let groups: LabeledValue<string>[] = await answer.json();
  if (addGroupKindSuffixToLabel) {
    const groupKindSuffix = i18next.t(`ListView.searchForm.${groupKind}`);
    groups = groups.map(group => ({
      label: `${group.label} (${groupKindSuffix})`,
      value: group.value,
    }));
  }
  return groups;
};

export const fetchUser = async(username: string): Promise<UserDetailResponseModel | null> => {
  if (needMock) {
    return fetchMockData(userDetailResponseModel, 'fetchUser');
  }

  const answer = await fetchAuthenticated(`/ucsschool/bff-users/v1/users/${username}`);
  if (answer.status === 404) {
    return null;
  }
  return await answer.json();
};


export type SaveUserError = {
  type: 'generic';
  message: string;
} | {
  type: 'fieldErrors';
  errors: {
    field: string;
    message: string;
  }[];
};

const getSaveUserError = async(response: Response, context: 'updateUser' | 'createUser'): Promise<SaveUserError | null> => {
  const errorJson = await getError(response);
  if (!errorJson) {
    return null;
  }
  if (response.status === 422) {
    // FIXME validate against json.schema
    if (Array.isArray(errorJson.detail)) {
      const detail: {loc: string[]; msg: string; type: string}[] = errorJson.detail;
      const error: SaveUserError = {
        type: 'fieldErrors',
        errors: [],
      };
      for (const iDetail of detail) {
        const field = iDetail.loc[iDetail.loc.length - 1];
        if (field) {
          error.errors.push({
            field,
            message: iDetail.msg,
          });
        }
      }
      return error;
    }
  }
  if (response.status === 404) {
    return {
      type: 'generic',
      message: i18next.t(`dataAccess.${context}.error.404`),
    };
  }
  return {
    type: 'generic',
    message: String(errorJson.detail),
  };
};
export const updateUser = async(url: string, values: FormValues): Promise<SaveUserError | null> => {
  if (needMock) {
    /*
    return fetchMockData({
      type: 'fieldErrors',
      errors: [
        {
          field: 'password',
          message: 'This should be overwritten in EditView',
        },
      ],
    }, 'updateUser');
    */
    return fetchMockData(null, 'updateUser');
  }

  const fixedProtocolUrl = url.replace(/https?:/, location.protocol);
  const answer = await fetchAuthenticated(`${fixedProtocolUrl}`, {
    method: 'PATCH',
    headers: {
      'Content-Type': 'application/json',
    },
    body: JSON.stringify(values),
  });
  return await getSaveUserError(answer, 'updateUser');
};


interface CreateUserUser {
  role: string;
  name: string;
  email?: string;
}
interface CreateUserSuccess {
  status: 'success';
  user: CreateUserUser;
}
interface CreateUserError {
  status: 'error';
  error: SaveUserError;
}
type CreateUserResponse = CreateUserSuccess | CreateUserError;
export const createUser = async(values: FormValues): Promise<CreateUserResponse> => {
  if (needMock) {
    return {
      status: 'success',
      user: await fetchMockData({
        name: 'user1',
        role: 'Student',
      }, 'createUser'),
    };
  }
  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/users/', {
    method: 'POST',
    headers: {
      'Content-Type': 'application/json',
    },
    body: JSON.stringify(values),
  });

  const error = await getSaveUserError(answer, 'createUser');
  if (error) {
    return {
      status: 'error',
      error,
    };
  }

  return {
    status: 'success',
    user: await answer.json(),
  };

  // This commented-out block helped me identify issues that were occuring
  //  during e2e-testing only (cypress doesn't show the response body of
  //  failed requests)
  //
  // if (!answer.ok) {
  //   await new Promise((finishedReading): void => {
  //     let body = '';
  //     if (!answer.body) {
  //       console.log('No response body');
  //     } else {
  //       console.log('Reading response body');
  //       const reader = answer.body.getReader();
  //       const processBodyChunk = ({done, value}: ReadableStreamDefaultReadResult<Uint8Array>): Promise<unknown> => {
  //         body += new TextDecoder().decode(value);
  //         if (done) {
  //           finishedReading(body);
  //           console.log(body);
  //           return Promise.all([]);
  //         }
  //         return reader.read().then(processBodyChunk);
  //       };
  //       reader.read().then(processBodyChunk);
  //     }
  //   });
  // }
};

export const deleteUsers = async(usernames: string[]): Promise<{id: string; error: string}[]> => {
  if (needMock) {
    return fetchMockData([
      {
        id: usernames[0] as string,
        error: 'some error',
      },
    ], 'deleteUsers');
  }
  const responses = await Promise.all(usernames.map(username => fetchAuthenticated(`/ucsschool/bff-users/v1/users/${username}`, {
    method: 'DELETE',
  })));
  const errors = await Promise.all(responses.map(x => getError(x)));
  const fails: {id: string; error: string}[] = [];
  for (let x = 0; x < responses.length; x++) {
    const error = errors[x];
    if (error) {
      fails.push({
        id: usernames[x] as string,
        error: JSON.stringify(error.detail),
      });
    }
  }
  return fails;
};

interface CsvResponseModelSuccess {
  status: 'success';
  blob: Blob;
  filename: string;
}

export interface CsvResponseModelError {
  status: 'error';
  type: '422' | 'other';
  message: string;
}

export type CsvResponseModel = CsvResponseModelError | CsvResponseModelSuccess;

export const exportUsers = async(usernames: string[]): Promise<CsvResponseModel> => {
  if (needMock) {
    return fetchMockData(csvResponseModel, 'exportUsers');
  }

  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/csv', {
    method: 'POST',
    headers: {
      'accept': 'text/csv',
      'Content-Type': 'application/json',
    },
    body: JSON.stringify({users: usernames}),
  });

  const error = await getError(answer);
  if (error) {
    return {
      status: 'error',
      type: answer.status === 422 ? '422' : 'other',
      message: JSON.stringify(error.detail),
    };
  }

  const content = await answer.blob();
  const filename = await answer.headers.get('Content-Disposition');
  const matches = filename && /filename\s*=\s*(?<filename>[\S\s]+)/g.exec(filename);
  const extractedName = matches && matches.length > 0 && matches[1];

  return {
    status: 'success',
    blob: content,
    filename: extractedName ? extractedName : 'user-export.csv',
  };
};


interface BackendSetPasswordsResponse {
  failedUsers: {id: string; error: string}[];
  file: {
    filename: string;
    data: string;
  };
}
interface SetPasswordResponseModelSuccess {
  status: 'success';
  response: BackendSetPasswordsResponse;
}
export interface SetPasswordResponseModelError {
  status: 'error';
}
export type SetPasswordResponseModel = SetPasswordResponseModelError | SetPasswordResponseModelSuccess;
export const setPasswords = async(password: string, usernames: string[]): Promise<SetPasswordResponseModel> => {
  if (needMock) {
    /*
    return fetchMockData({
      status: 'error',
    }, 'setPasswords');
    */
    return fetchMockData(setPasswordResponseModel, 'setPasswords');
  }

  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/set-passwords', {
    method: 'POST',
    headers: {
      'accept': 'application/json',
      'Content-Type': 'application/json',
    },
    body: JSON.stringify({
      users: usernames,
      password,
    }),
  });

  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.setPasswords.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }

  const response: BackendSetPasswordsResponse = await answer.json();
  return {
    status: 'success',
    response,
  };
};

type BackendSetWifiPasswordsResponse = BackendSetPasswordsResponse;
interface SetWifiPasswordResponseModelSuccess extends SetPasswordResponseModelSuccess {
  response: BackendSetWifiPasswordsResponse;
}
type SetWifiPasswordResponseModelError = SetPasswordResponseModelError;
export type SetWifiPasswordResponseModel = SetWifiPasswordResponseModelError | SetWifiPasswordResponseModelSuccess;
export const setWifiPasswords = async(password: string, usernames: string[]): Promise<SetWifiPasswordResponseModel> => {
  if (needMock) {
    return fetchMockData(setWifiPasswordResponseModel, 'setWifiPasswords');
  }

  // TODO please replace with the previously unknown correct route
  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/set-wifi-passwords', {
    method: 'POST',
    headers: {
      'accept': 'application/json',
      'Content-Type': 'application/json',
    },
    body: JSON.stringify({
      users: usernames,
      password,
    }),
  });

  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.setWifiPasswords.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }

  const response: BackendSetWifiPasswordsResponse = await answer.json();
  return {
    status: 'success',
    response,
  };
};


export const toggleUserActivationStates = async(usernames: string[], action: ToggleActiveAction): Promise<{id: string; error: string}[]> => {
  if (needMock) {
    return fetchMockData([
      {
        id: usernames[0] as string,
        error: 'some error',
      },
    ], 'toggleUserActivationStates');
  }

  const responses = await Promise.all(usernames.map(username => fetchAuthenticated(`/ucsschool/bff-users/v1/users/${username}/${action}`, {
    method: 'PATCH',
  })));
  const errors = await Promise.all(responses.map(x => getError(x)));
  const fails: {id: string; error: string}[] = [];
  for (let x = 0; x < responses.length; x++) {
    const error = errors[x];
    if (error) {
      fails.push({
        id: usernames[x] as string,
        error: JSON.stringify(error.detail),
      });
    }
  }
  return fails;
};

export const toggleWifiActivationStates = async(usernames: string[], action: ToggleWifiActiveAction): Promise<{id: string; error: string}[]> => {
  if (needMock) {
    return fetchMockData([
      {
        id: usernames[0] as string,
        error: 'some error',
      },
    ], 'toggleWifiActivationStates');
  }

  const responses = await Promise.all(usernames.map(username => fetchAuthenticated(`/ucsschool/bff-users/v1/wifi/${username}/${action}`, {
    method: 'PATCH',
  })));
  const errors = await Promise.all(responses.map(x => getError(x)));
  const fails: {id: string; error: string}[] = [];
  for (let x = 0; x < responses.length; x++) {
    const error = errors[x];
    if (error) {
      fails.push({
        id: usernames[x] as string,
        error: JSON.stringify(error.detail),
      });
    }
  }
  return fails;
};
